from pgcli.packages.sqlcompletion import (
    suggest_type,
    Special,
    Database,
    Schema,
    Table,
    Column,
    View,
    Keyword,
    FromClauseItem,
    Function,
    Datatype,
    Alias,
    JoinCondition,
    Join,
)
from pgcli.packages.parseutils.tables import TableReference
import pytest


def cols_etc(table, schema=None, alias=None, is_function=False, parent=None, last_keyword=None):
    """Returns the expected select-clause suggestions for a single-table
    select."""
    return {
        Column(
            table_refs=(TableReference(schema, table, alias, is_function),),
            qualifiable=True,
        ),
        Function(schema=parent),
        Keyword(last_keyword),
    }


def test_select_suggests_cols_with_visible_table_scope():
    suggestions = suggest_type("SELECT  FROM tabl", "SELECT ")
    assert set(suggestions) == cols_etc("tabl", last_keyword="SELECT")


def test_select_suggests_cols_with_qualified_table_scope():
    suggestions = suggest_type("SELECT  FROM sch.tabl", "SELECT ")
    assert set(suggestions) == cols_etc("tabl", "sch", last_keyword="SELECT")


def test_cte_does_not_crash():
    sql = "WITH CTE AS (SELECT F.* FROM Foo F WHERE F.Bar > 23) SELECT C.* FROM CTE C WHERE C.FooID BETWEEN 123 AND 234;"
    for i in range(len(sql)):
        suggest_type(sql[: i + 1], sql[: i + 1])


@pytest.mark.parametrize("expression", ['SELECT * FROM "tabl" WHERE '])
def test_where_suggests_columns_functions_quoted_table(expression):
    expected = cols_etc("tabl", alias='"tabl"', last_keyword="WHERE")
    suggestions = suggest_type(expression, expression)
    assert expected == set(suggestions)


@pytest.mark.parametrize(
    "expression",
    [
        "INSERT INTO OtherTabl(ID, Name) SELECT * FROM tabl WHERE ",
        "INSERT INTO OtherTabl SELECT * FROM tabl WHERE ",
        "SELECT * FROM tabl WHERE ",
        "SELECT * FROM tabl WHERE (",
        "SELECT * FROM tabl WHERE foo = ",
        "SELECT * FROM tabl WHERE bar OR ",
        "SELECT * FROM tabl WHERE foo = 1 AND ",
        "SELECT * FROM tabl WHERE (bar > 10 AND ",
        "SELECT * FROM tabl WHERE (bar AND (baz OR (qux AND (",
        "SELECT * FROM tabl WHERE 10 < ",
        "SELECT * FROM tabl WHERE foo BETWEEN ",
        "SELECT * FROM tabl WHERE foo BETWEEN foo AND ",
    ],
)
def test_where_suggests_columns_functions(expression):
    suggestions = suggest_type(expression, expression)
    assert set(suggestions) == cols_etc("tabl", last_keyword="WHERE")


@pytest.mark.parametrize(
    "expression",
    ["SELECT * FROM tabl WHERE foo IN (", "SELECT * FROM tabl WHERE foo IN (bar, "],
)
def test_where_in_suggests_columns(expression):
    suggestions = suggest_type(expression, expression)
    assert set(suggestions) == cols_etc("tabl", last_keyword="WHERE")


@pytest.mark.parametrize("expression", ["SELECT 1 AS ", "SELECT 1 FROM tabl AS "])
def test_after_as(expression):
    suggestions = suggest_type(expression, expression)
    assert set(suggestions) == set()


def test_where_equals_any_suggests_columns_or_keywords():
    text = "SELECT * FROM tabl WHERE foo = ANY("
    suggestions = suggest_type(text, text)
    assert set(suggestions) == cols_etc("tabl", last_keyword="WHERE")


def test_lparen_suggests_cols_and_funcs():
    suggestion = suggest_type("SELECT MAX( FROM tbl", "SELECT MAX(")
    assert set(suggestion) == {
        Column(table_refs=((None, "tbl", None, False),), qualifiable=True),
        Function(schema=None),
        Keyword("("),
    }


def test_select_suggests_cols_and_funcs():
    suggestions = suggest_type("SELECT ", "SELECT ")
    assert set(suggestions) == {
        Column(table_refs=(), qualifiable=True),
        Function(schema=None),
        Keyword("SELECT"),
    }


@pytest.mark.parametrize("expression", ["INSERT INTO ", "COPY ", "UPDATE ", "DESCRIBE "])
def test_suggests_tables_views_and_schemas(expression):
    suggestions = suggest_type(expression, expression)
    assert set(suggestions) == {Table(schema=None), View(schema=None), Schema()}


@pytest.mark.parametrize("expression", ["SELECT * FROM "])
def test_suggest_tables_views_schemas_and_functions(expression):
    suggestions = suggest_type(expression, expression)
    assert set(suggestions) == {FromClauseItem(schema=None), Schema()}


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM foo JOIN bar on bar.barid = foo.barid JOIN ",
        "SELECT * FROM foo JOIN bar USING (barid) JOIN ",
    ],
)
def test_suggest_after_join_with_two_tables(expression):
    suggestions = suggest_type(expression, expression)
    tables = ((None, "foo", None, False), (None, "bar", None, False))
    assert set(suggestions) == {
        FromClauseItem(schema=None, table_refs=tables),
        Join(tables, None),
        Schema(),
    }


@pytest.mark.parametrize("expression", ["SELECT * FROM foo JOIN ", "SELECT * FROM foo JOIN bar"])
def test_suggest_after_join_with_one_table(expression):
    suggestions = suggest_type(expression, expression)
    tables = ((None, "foo", None, False),)
    assert set(suggestions) == {
        FromClauseItem(schema=None, table_refs=tables),
        Join(((None, "foo", None, False),), None),
        Schema(),
    }


@pytest.mark.parametrize("expression", ["INSERT INTO sch.", "COPY sch.", "DESCRIBE sch."])
def test_suggest_qualified_tables_and_views(expression):
    suggestions = suggest_type(expression, expression)
    assert set(suggestions) == {Table(schema="sch"), View(schema="sch")}


@pytest.mark.parametrize("expression", ["UPDATE sch."])
def test_suggest_qualified_aliasable_tables_and_views(expression):
    suggestions = suggest_type(expression, expression)
    assert set(suggestions) == {Table(schema="sch"), View(schema="sch")}


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM sch.",
        'SELECT * FROM sch."',
        'SELECT * FROM sch."foo',
        'SELECT * FROM "sch".',
        'SELECT * FROM "sch"."',
    ],
)
def test_suggest_qualified_tables_views_and_functions(expression):
    suggestions = suggest_type(expression, expression)
    assert set(suggestions) == {FromClauseItem(schema="sch")}


@pytest.mark.parametrize("expression", ["SELECT * FROM foo JOIN sch."])
def test_suggest_qualified_tables_views_functions_and_joins(expression):
    suggestions = suggest_type(expression, expression)
    tbls = ((None, "foo", None, False),)
    assert set(suggestions) == {
        FromClauseItem(schema="sch", table_refs=tbls),
        Join(tbls, "sch"),
    }


def test_truncate_suggests_tables_and_schemas():
    suggestions = suggest_type("TRUNCATE ", "TRUNCATE ")
    assert set(suggestions) == {Table(schema=None), Schema()}


def test_truncate_suggests_qualified_tables():
    suggestions = suggest_type("TRUNCATE sch.", "TRUNCATE sch.")
    assert set(suggestions) == {Table(schema="sch")}


@pytest.mark.parametrize("text", ["SELECT DISTINCT ", "INSERT INTO foo SELECT DISTINCT "])
def test_distinct_suggests_cols(text):
    suggestions = suggest_type(text, text)
    assert set(suggestions) == {
        Column(table_refs=(), local_tables=(), qualifiable=True),
        Function(schema=None),
        Keyword("DISTINCT"),
    }


@pytest.mark.parametrize(
    "text, text_before, last_keyword",
    [
        ("SELECT DISTINCT FROM tbl x JOIN tbl1 y", "SELECT DISTINCT", "SELECT"),
        (
            "SELECT * FROM tbl x JOIN tbl1 y ORDER BY ",
            "SELECT * FROM tbl x JOIN tbl1 y ORDER BY ",
            "ORDER BY",
        ),
        (
            "SELECT * FROM tbl x JOIN tbl1 y GROUP BY ",
            "SELECT * FROM tbl x JOIN tbl1 y GROUP BY ",
            "GROUP BY",
        ),
    ],
)
def test_distinct_and_order_by_suggestions_with_aliases(text, text_before, last_keyword):
    suggestions = suggest_type(text, text_before)
    assert set(suggestions) == {
        Column(
            table_refs=(
                TableReference(None, "tbl", "x", False),
                TableReference(None, "tbl1", "y", False),
            ),
            local_tables=(),
            qualifiable=True,
        ),
        Function(schema=None),
        Keyword(last_keyword),
    }


@pytest.mark.parametrize(
    "text, text_before",
    [
        ("SELECT DISTINCT x. FROM tbl x JOIN tbl1 y", "SELECT DISTINCT x."),
        (
            "SELECT * FROM tbl x JOIN tbl1 y ORDER BY x.",
            "SELECT * FROM tbl x JOIN tbl1 y ORDER BY x.",
        ),
        (
            "SELECT * FROM tbl x JOIN tbl1 y GROUP BY x.",
            "SELECT * FROM tbl x JOIN tbl1 y GROUP BY x.",
        ),
    ],
)
def test_distinct_and_order_by_suggestions_with_alias_given(text, text_before):
    suggestions = suggest_type(text, text_before)
    assert set(suggestions) == {
        Column(
            table_refs=(TableReference(None, "tbl", "x", False),),
            local_tables=(),
            qualifiable=False,
        ),
        Table(schema="x"),
        View(schema="x"),
        Function(schema="x"),
    }


def test_function_arguments_with_alias_given():
    suggestions = suggest_type("SELECT avg(x. FROM tbl x, tbl2 y", "SELECT avg(x.")

    assert set(suggestions) == {
        Column(
            table_refs=(TableReference(None, "tbl", "x", False),),
            local_tables=(),
            qualifiable=False,
        ),
        Table(schema="x"),
        View(schema="x"),
        Function(schema="x"),
    }


def test_col_comma_suggests_cols():
    suggestions = suggest_type("SELECT a, b, FROM tbl", "SELECT a, b,")
    assert set(suggestions) == {
        Column(table_refs=((None, "tbl", None, False),), qualifiable=True),
        Function(schema=None),
        Keyword("SELECT"),
    }


def test_table_comma_suggests_tables_and_schemas():
    suggestions = suggest_type("SELECT a, b FROM tbl1, ", "SELECT a, b FROM tbl1, ")
    assert set(suggestions) == {FromClauseItem(schema=None), Schema()}


def test_into_suggests_tables_and_schemas():
    suggestion = suggest_type("INSERT INTO ", "INSERT INTO ")
    assert set(suggestion) == {Table(schema=None), View(schema=None), Schema()}


@pytest.mark.parametrize("text", ["INSERT INTO abc (", "INSERT INTO abc () SELECT * FROM hij;"])
def test_insert_into_lparen_suggests_cols(text):
    suggestions = suggest_type(text, "INSERT INTO abc (")
    assert suggestions == (Column(table_refs=((None, "abc", None, False),), context="insert"),)


def test_insert_into_lparen_partial_text_suggests_cols():
    suggestions = suggest_type("INSERT INTO abc (i", "INSERT INTO abc (i")
    assert suggestions == (Column(table_refs=((None, "abc", None, False),), context="insert"),)


def test_insert_into_lparen_comma_suggests_cols():
    suggestions = suggest_type("INSERT INTO abc (id,", "INSERT INTO abc (id,")
    assert suggestions == (Column(table_refs=((None, "abc", None, False),), context="insert"),)


def test_partially_typed_col_name_suggests_col_names():
    suggestions = suggest_type("SELECT * FROM tabl WHERE col_n", "SELECT * FROM tabl WHERE col_n")
    assert set(suggestions) == cols_etc("tabl", last_keyword="WHERE")


def test_dot_suggests_cols_of_a_table_or_schema_qualified_table():
    suggestions = suggest_type("SELECT tabl. FROM tabl", "SELECT tabl.")
    assert set(suggestions) == {
        Column(table_refs=((None, "tabl", None, False),)),
        Table(schema="tabl"),
        View(schema="tabl"),
        Function(schema="tabl"),
    }


@pytest.mark.parametrize(
    "sql",
    [
        "SELECT t1. FROM tabl1 t1",
        "SELECT t1. FROM tabl1 t1, tabl2 t2",
        'SELECT t1. FROM "tabl1" t1',
        'SELECT t1. FROM "tabl1" t1, "tabl2" t2',
    ],
)
def test_dot_suggests_cols_of_an_alias(sql):
    suggestions = suggest_type(sql, "SELECT t1.")
    assert set(suggestions) == {
        Table(schema="t1"),
        View(schema="t1"),
        Column(table_refs=((None, "tabl1", "t1", False),)),
        Function(schema="t1"),
    }


@pytest.mark.parametrize(
    "sql",
    [
        "SELECT * FROM tabl1 t1 WHERE t1.",
        "SELECT * FROM tabl1 t1, tabl2 t2 WHERE t1.",
        'SELECT * FROM "tabl1" t1 WHERE t1.',
        'SELECT * FROM "tabl1" t1, tabl2 t2 WHERE t1.',
    ],
)
def test_dot_suggests_cols_of_an_alias_where(sql):
    suggestions = suggest_type(sql, sql)
    assert set(suggestions) == {
        Table(schema="t1"),
        View(schema="t1"),
        Column(table_refs=((None, "tabl1", "t1", False),)),
        Function(schema="t1"),
    }


def test_dot_col_comma_suggests_cols_or_schema_qualified_table():
    suggestions = suggest_type("SELECT t1.a, t2. FROM tabl1 t1, tabl2 t2", "SELECT t1.a, t2.")
    assert set(suggestions) == {
        Column(table_refs=((None, "tabl2", "t2", False),)),
        Table(schema="t2"),
        View(schema="t2"),
        Function(schema="t2"),
    }


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM (",
        "SELECT * FROM foo WHERE EXISTS (",
        "SELECT * FROM foo WHERE bar AND NOT EXISTS (",
    ],
)
def test_sub_select_suggests_keyword(expression):
    suggestion = suggest_type(expression, expression)
    assert suggestion == (Keyword(),)


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM (S",
        "SELECT * FROM foo WHERE EXISTS (S",
        "SELECT * FROM foo WHERE bar AND NOT EXISTS (S",
    ],
)
def test_sub_select_partial_text_suggests_keyword(expression):
    suggestion = suggest_type(expression, expression)
    assert suggestion == (Keyword(),)


def test_outer_table_reference_in_exists_subquery_suggests_columns():
    q = "SELECT * FROM foo f WHERE EXISTS (SELECT 1 FROM bar WHERE f."
    suggestions = suggest_type(q, q)
    assert set(suggestions) == {
        Column(table_refs=((None, "foo", "f", False),)),
        Table(schema="f"),
        View(schema="f"),
        Function(schema="f"),
    }


@pytest.mark.parametrize("expression", ["SELECT * FROM (SELECT * FROM "])
def test_sub_select_table_name_completion(expression):
    suggestion = suggest_type(expression, expression)
    assert set(suggestion) == {FromClauseItem(schema=None), Schema()}


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM foo WHERE EXISTS (SELECT * FROM ",
        "SELECT * FROM foo WHERE bar AND NOT EXISTS (SELECT * FROM ",
    ],
)
def test_sub_select_table_name_completion_with_outer_table(expression):
    suggestion = suggest_type(expression, expression)
    tbls = ((None, "foo", None, False),)
    assert set(suggestion) == {FromClauseItem(schema=None, table_refs=tbls), Schema()}


def test_sub_select_col_name_completion():
    suggestions = suggest_type("SELECT * FROM (SELECT  FROM abc", "SELECT * FROM (SELECT ")
    assert set(suggestions) == {
        Column(table_refs=((None, "abc", None, False),), qualifiable=True),
        Function(schema=None),
        Keyword("SELECT"),
    }


@pytest.mark.xfail
def test_sub_select_multiple_col_name_completion():
    suggestions = suggest_type("SELECT * FROM (SELECT a, FROM abc", "SELECT * FROM (SELECT a, ")
    assert set(suggestions) == cols_etc("abc")


def test_sub_select_dot_col_name_completion():
    suggestions = suggest_type("SELECT * FROM (SELECT t. FROM tabl t", "SELECT * FROM (SELECT t.")
    assert set(suggestions) == {
        Column(table_refs=((None, "tabl", "t", False),)),
        Table(schema="t"),
        View(schema="t"),
        Function(schema="t"),
    }


@pytest.mark.parametrize("join_type", ("", "INNER", "LEFT", "RIGHT OUTER"))
@pytest.mark.parametrize("tbl_alias", ("", "foo"))
def test_join_suggests_tables_and_schemas(tbl_alias, join_type):
    text = f"SELECT * FROM abc {tbl_alias} {join_type} JOIN "
    suggestion = suggest_type(text, text)
    tbls = ((None, "abc", tbl_alias or None, False),)
    assert set(suggestion) == {
        FromClauseItem(schema=None, table_refs=tbls),
        Schema(),
        Join(tbls, None),
    }


def test_left_join_with_comma():
    text = "select * from foo f left join bar b,"
    suggestions = suggest_type(text, text)
    # tbls should also include (None, 'bar', 'b', False)
    # but there's a bug with commas
    tbls = ((None, "foo", "f", False),)
    assert set(suggestions) == {FromClauseItem(schema=None, table_refs=tbls), Schema()}


@pytest.mark.parametrize(
    "sql",
    [
        "SELECT * FROM abc a JOIN def d ON a.",
        "SELECT * FROM abc a JOIN def d ON a.id = d.id AND a.",
    ],
)
def test_join_alias_dot_suggests_cols1(sql):
    suggestions = suggest_type(sql, sql)
    tables = ((None, "abc", "a", False), (None, "def", "d", False))
    assert set(suggestions) == {
        Column(table_refs=((None, "abc", "a", False),)),
        Table(schema="a"),
        View(schema="a"),
        Function(schema="a"),
        JoinCondition(table_refs=tables, parent=(None, "abc", "a", False)),
    }


@pytest.mark.parametrize(
    "sql",
    [
        "SELECT * FROM abc a JOIN def d ON a.id = d.",
        "SELECT * FROM abc a JOIN def d ON a.id = d.id AND a.id2 = d.",
    ],
)
def test_join_alias_dot_suggests_cols2(sql):
    suggestion = suggest_type(sql, sql)
    assert set(suggestion) == {
        Column(table_refs=((None, "def", "d", False),)),
        Table(schema="d"),
        View(schema="d"),
        Function(schema="d"),
    }


@pytest.mark.parametrize(
    "sql",
    [
        "select a.x, b.y from abc a join bcd b on ",
        """select a.x, b.y
from abc a
join bcd b on
""",
        """select a.x, b.y
from abc a
join bcd b
on """,
        "select a.x, b.y from abc a join bcd b on a.id = b.id OR ",
    ],
)
def test_on_suggests_aliases_and_join_conditions(sql):
    suggestions = suggest_type(sql, sql)
    tables = ((None, "abc", "a", False), (None, "bcd", "b", False))
    assert set(suggestions) == {
        JoinCondition(table_refs=tables, parent=None),
        Alias(aliases=("a", "b")),
    }


@pytest.mark.parametrize(
    "sql",
    [
        "select abc.x, bcd.y from abc join bcd on abc.id = bcd.id AND ",
        "select abc.x, bcd.y from abc join bcd on ",
    ],
)
def test_on_suggests_tables_and_join_conditions(sql):
    suggestions = suggest_type(sql, sql)
    tables = ((None, "abc", None, False), (None, "bcd", None, False))
    assert set(suggestions) == {
        JoinCondition(table_refs=tables, parent=None),
        Alias(aliases=("abc", "bcd")),
    }


@pytest.mark.parametrize(
    "sql",
    [
        "select a.x, b.y from abc a join bcd b on a.id = ",
        "select a.x, b.y from abc a join bcd b on a.id = b.id AND a.id2 = ",
    ],
)
def test_on_suggests_aliases_right_side(sql):
    suggestions = suggest_type(sql, sql)
    assert suggestions == (Alias(aliases=("a", "b")),)


@pytest.mark.parametrize(
    "sql",
    [
        "select abc.x, bcd.y from abc join bcd on abc.id = bcd.id and ",
        "select abc.x, bcd.y from abc join bcd on ",
    ],
)
def test_on_suggests_tables_and_join_conditions_right_side(sql):
    suggestions = suggest_type(sql, sql)
    tables = ((None, "abc", None, False), (None, "bcd", None, False))
    assert set(suggestions) == {
        JoinCondition(table_refs=tables, parent=None),
        Alias(aliases=("abc", "bcd")),
    }


@pytest.mark.parametrize(
    "text",
    (
        "select * from abc inner join def using (",
        "select * from abc inner join def using (col1, ",
        "insert into hij select * from abc inner join def using (",
        """insert into hij(x, y, z)
    select * from abc inner join def using (col1, """,
        """insert into hij (a,b,c)
    select * from abc inner join def using (col1, """,
    ),
)
def test_join_using_suggests_common_columns(text):
    tables = ((None, "abc", None, False), (None, "def", None, False))
    assert set(suggest_type(text, text)) == {Column(table_refs=tables, require_last_table=True)}


def test_suggest_columns_after_multiple_joins():
    sql = """select * from t1
            inner join t2 ON
              t1.id = t2.t1_id
            inner join t3 ON
              t2.id = t3."""
    suggestions = suggest_type(sql, sql)
    assert Column(table_refs=((None, "t3", None, False),)) in set(suggestions)


def test_2_statements_2nd_current():
    suggestions = suggest_type("select * from a; select * from ", "select * from a; select * from ")
    assert set(suggestions) == {FromClauseItem(schema=None), Schema()}

    suggestions = suggest_type("select * from a; select  from b", "select * from a; select ")
    assert set(suggestions) == {
        Column(table_refs=((None, "b", None, False),), qualifiable=True),
        Function(schema=None),
        Keyword("SELECT"),
    }

    # Should work even if first statement is invalid
    suggestions = suggest_type("select * from; select * from ", "select * from; select * from ")
    assert set(suggestions) == {FromClauseItem(schema=None), Schema()}


def test_2_statements_1st_current():
    suggestions = suggest_type("select * from ; select * from b", "select * from ")
    assert set(suggestions) == {FromClauseItem(schema=None), Schema()}

    suggestions = suggest_type("select  from a; select * from b", "select ")
    assert set(suggestions) == cols_etc("a", last_keyword="SELECT")


def test_3_statements_2nd_current():
    suggestions = suggest_type(
        "select * from a; select * from ; select * from c",
        "select * from a; select * from ",
    )
    assert set(suggestions) == {FromClauseItem(schema=None), Schema()}

    suggestions = suggest_type("select * from a; select  from b; select * from c", "select * from a; select ")
    assert set(suggestions) == cols_etc("b", last_keyword="SELECT")


@pytest.mark.parametrize(
    "text",
    [
        """
CREATE OR REPLACE FUNCTION func() RETURNS setof int AS $$
SELECT  FROM foo;
SELECT 2 FROM bar;
$$ language sql;
    """,
        """create function func2(int, varchar)
RETURNS text
language sql AS
$func$
SELECT 2 FROM bar;
SELECT  FROM foo;
$func$
    """,
        """
CREATE OR REPLACE FUNCTION func() RETURNS setof int AS $func$
SELECT 3 FROM foo;
SELECT 2 FROM bar;
$$ language sql;
create function func2(int, varchar)
RETURNS text
language sql AS
$func$
SELECT 2 FROM bar;
SELECT  FROM foo;
$func$
    """,
        """
SELECT * FROM baz;
CREATE OR REPLACE FUNCTION func() RETURNS setof int AS $func$
SELECT  FROM foo;
SELECT 2 FROM bar;
$$ language sql;
create function func2(int, varchar)
RETURNS text
language sql AS
$func$
SELECT 3 FROM bar;
SELECT  FROM foo;
$func$
SELECT * FROM qux;
    """,
    ],
)
def test_statements_in_function_body(text):
    suggestions = suggest_type(text, text[: text.find("  ") + 1])
    assert set(suggestions) == {
        Column(table_refs=((None, "foo", None, False),), qualifiable=True),
        Function(schema=None),
        Keyword("SELECT"),
    }


functions = [
    """
CREATE OR REPLACE FUNCTION func() RETURNS setof int AS $$
SELECT 1 FROM foo;
SELECT 2 FROM bar;
$$ language sql;
    """,
    """
create function func2(int, varchar)
RETURNS text
language sql AS
'
SELECT 2 FROM bar;
SELECT 1 FROM foo;
';
    """,
]


@pytest.mark.parametrize("text", functions)
def test_statements_with_cursor_after_function_body(text):
    suggestions = suggest_type(text, text[: text.find("; ") + 1])
    assert set(suggestions) == {Keyword(), Special()}


@pytest.mark.parametrize("text", functions)
def test_statements_with_cursor_before_function_body(text):
    suggestions = suggest_type(text, "")
    assert set(suggestions) == {Keyword(), Special()}


def test_create_db_with_template():
    suggestions = suggest_type("create database foo with template ", "create database foo with template ")

    assert set(suggestions) == {Database()}


@pytest.mark.parametrize("initial_text", ("", "    ", "\t \t", "\n"))
def test_specials_included_for_initial_completion(initial_text):
    suggestions = suggest_type(initial_text, initial_text)

    assert set(suggestions) == {Keyword(), Special()}


def test_drop_schema_qualified_table_suggests_only_tables():
    text = "DROP TABLE schema_name.table_name"
    suggestions = suggest_type(text, text)
    assert suggestions == (Table(schema="schema_name"),)


@pytest.mark.parametrize("text", (",", "  ,", "sel ,"))
def test_handle_pre_completion_comma_gracefully(text):
    suggestions = suggest_type(text, text)

    assert iter(suggestions)


def test_drop_schema_suggests_schemas():
    sql = "DROP SCHEMA "
    assert suggest_type(sql, sql) == (Schema(),)


@pytest.mark.parametrize("text", ["SELECT x::", "SELECT x::y", "SELECT (x + y)::"])
def test_cast_operator_suggests_types(text):
    assert set(suggest_type(text, text)) == {
        Datatype(schema=None),
        Table(schema=None),
        Schema(),
    }


@pytest.mark.parametrize("text", ["SELECT foo::bar.", "SELECT foo::bar.baz", "SELECT (x + y)::bar."])
def test_cast_operator_suggests_schema_qualified_types(text):
    assert set(suggest_type(text, text)) == {
        Datatype(schema="bar"),
        Table(schema="bar"),
    }


def test_alter_column_type_suggests_types():
    q = "ALTER TABLE foo ALTER COLUMN bar TYPE "
    assert set(suggest_type(q, q)) == {
        Datatype(schema=None),
        Table(schema=None),
        Schema(),
    }


@pytest.mark.parametrize(
    "text",
    [
        "SELECT type ",
        "SELECT type, ",
        "SELECT id, type, ",
    ],
)
def test_column_named_type_still_suggests_columns(text):
    # `type` is a non-reserved word, so sqlparse tags a column literally
    # named "type" as the TYPE keyword; it must not switch the SELECT list
    # over to datatype suggestions.
    suggestions = suggest_type(text, text)
    assert Column in {type(s) for s in suggestions}
    assert Datatype(schema=None) not in suggestions


@pytest.mark.parametrize(
    "text",
    [
        "CREATE TABLE foo (bar ",
        "CREATE TABLE foo (bar DOU",
        "CREATE TABLE foo (bar INT, baz ",
        "CREATE TABLE foo (bar INT, baz TEXT, qux ",
        "CREATE FUNCTION foo (bar ",
        "CREATE FUNCTION foo (bar INT, baz ",
        "SELECT * FROM foo() AS bar (baz ",
        "SELECT * FROM foo() AS bar (baz INT, qux ",
        # make sure this doesn't trigger special completion
        "CREATE TABLE foo (dt d",
    ],
)
def test_identifier_suggests_types_in_parentheses(text):
    assert set(suggest_type(text, text)) == {
        Datatype(schema=None),
        Table(schema=None),
        Schema(),
    }


@pytest.mark.parametrize(
    "text",
    [
        "SELECT foo ",
        "SELECT foo FROM bar ",
        "SELECT foo AS bar ",
        "SELECT foo bar ",
        "SELECT * FROM foo AS bar ",
        "SELECT * FROM foo bar ",
        "SELECT foo FROM (SELECT bar ",
    ],
)
def test_alias_suggests_keywords(text):
    suggestions = suggest_type(text, text)
    assert suggestions == (Keyword(),)


def test_invalid_sql():
    # issue 317
    text = "selt *"
    suggestions = suggest_type(text, text)
    assert suggestions == (Keyword(),)


@pytest.mark.parametrize(
    "text",
    ["SELECT * FROM foo where created > now() - ", "select * from foo where bar "],
)
def test_suggest_where_keyword(text):
    # https://github.com/dbcli/mycli/issues/135
    suggestions = suggest_type(text, text)
    assert set(suggestions) == cols_etc("foo", last_keyword="WHERE")


@pytest.mark.parametrize(
    "text, before, expected",
    [
        (
            "\\ns abc SELECT ",
            "SELECT ",
            [
                Column(table_refs=(), qualifiable=True),
                Function(schema=None),
                Keyword("SELECT"),
            ],
        ),
        ("\\ns abc SELECT foo ", "SELECT foo ", (Keyword(),)),
        (
            "\\ns abc SELECT t1. FROM tabl1 t1",
            "SELECT t1.",
            [
                Table(schema="t1"),
                View(schema="t1"),
                Column(table_refs=((None, "tabl1", "t1", False),)),
                Function(schema="t1"),
            ],
        ),
    ],
)
def test_named_query_completion(text, before, expected):
    suggestions = suggest_type(text, before)
    assert set(expected) == set(suggestions)


def test_select_suggests_fields_from_function():
    suggestions = suggest_type("SELECT  FROM func()", "SELECT ")
    assert set(suggestions) == cols_etc("func", is_function=True, last_keyword="SELECT")


@pytest.mark.parametrize("sql", ["("])
def test_leading_parenthesis(sql):
    # No assertion for now; just make sure it doesn't crash
    suggest_type(sql, sql)


@pytest.mark.parametrize("sql", ['select * from "', 'select * from "foo'])
def test_ignore_leading_double_quotes(sql):
    suggestions = suggest_type(sql, sql)
    assert FromClauseItem(schema=None) in set(suggestions)


@pytest.mark.parametrize(
    "sql",
    [
        "ALTER TABLE foo ALTER COLUMN ",
        "ALTER TABLE foo ALTER COLUMN bar",
        "ALTER TABLE foo DROP COLUMN ",
        "ALTER TABLE foo DROP COLUMN bar",
    ],
)
def test_column_keyword_suggests_columns(sql):
    suggestions = suggest_type(sql, sql)
    assert set(suggestions) == {Column(table_refs=((None, "foo", None, False),))}


def test_handle_unrecognized_kw_generously():
    sql = "SELECT * FROM sessions WHERE session = 1 AND "
    suggestions = suggest_type(sql, sql)
    expected = Column(table_refs=((None, "sessions", None, False),), qualifiable=True)

    assert expected in set(suggestions)


@pytest.mark.parametrize("sql", ["ALTER ", "ALTER TABLE foo ALTER "])
def test_keyword_after_alter(sql):
    assert Keyword("ALTER") in set(suggest_type(sql, sql))


def test_suggestion_when_setting_search_path():
    sql_set = "SET "
    suggestion_set = suggest_type(sql_set, sql_set)
    assert set(suggestion_set) == {Keyword("SET")}

    sql_set_search_path_to = "SET search_path TO "
    suggestion_set_search_path_to = suggest_type(sql_set_search_path_to, sql_set_search_path_to)
    assert set(suggestion_set_search_path_to) == {Schema()}



import click
from textwrap import dedent


keyring = None  # keyring will be loaded later


keyring_error_message = dedent(
    """\
    {}
    {}
    To remove this message do one of the following:
    - prepare keyring as described at: https://keyring.readthedocs.io/en/stable/
    - uninstall keyring: pip uninstall keyring
    - disable keyring in our configuration: add keyring = False to [main]"""
)


def keyring_initialize(keyring_enabled, *, logger):
    """Initialize keyring only if explicitly enabled"""
    global keyring

    if keyring_enabled:
        # Try best to load keyring (issue #1041).
        import importlib

        try:
            keyring = importlib.import_module("keyring")
        except ModuleNotFoundError as e:  # ImportError for Python 2, ModuleNotFoundError for Python 3
            logger.warning("import keyring failed: %r.", e)


def keyring_get_password(key):
    """Attempt to get password from keyring"""
    # Find password from store
    passwd = ""
    try:
        passwd = keyring.get_password("pgcli", key) or ""
    except Exception as e:
        click.secho(
            keyring_error_message.format("Load your password from keyring returned:", str(e)),
            err=True,
            fg="red",
        )
    return passwd


def keyring_set_password(key, passwd):
    try:
        keyring.set_password("pgcli", key, passwd)
    except Exception as e:
        click.secho(
            keyring_error_message.format("Set password in keyring returned:", str(e)),
            err=True,
            fg="red",
        )



import pytest
from pgcli.packages.sqlcompletion import (
    suggest_type,
    Special,
    Database,
    Schema,
    Table,
    View,
    Function,
    Datatype,
)


def test_slash_suggests_special():
    suggestions = suggest_type("\\", "\\")
    assert set(suggestions) == {Special()}


def test_slash_d_suggests_special():
    suggestions = suggest_type("\\d", "\\d")
    assert set(suggestions) == {Special()}


def test_dn_suggests_schemata():
    suggestions = suggest_type("\\dn ", "\\dn ")
    assert suggestions == (Schema(),)

    suggestions = suggest_type("\\dn xxx", "\\dn xxx")
    assert suggestions == (Schema(),)


def test_d_suggests_tables_views_and_schemas():
    suggestions = suggest_type(r"\d ", r"\d ")
    assert set(suggestions) == {Schema(), Table(schema=None), View(schema=None)}

    suggestions = suggest_type(r"\d xxx", r"\d xxx")
    assert set(suggestions) == {Schema(), Table(schema=None), View(schema=None)}


def test_d_dot_suggests_schema_qualified_tables_or_views():
    suggestions = suggest_type(r"\d myschema.", r"\d myschema.")
    assert set(suggestions) == {Table(schema="myschema"), View(schema="myschema")}

    suggestions = suggest_type(r"\d myschema.xxx", r"\d myschema.xxx")
    assert set(suggestions) == {Table(schema="myschema"), View(schema="myschema")}


def test_df_suggests_schema_or_function():
    suggestions = suggest_type("\\df xxx", "\\df xxx")
    assert set(suggestions) == {Function(schema=None, usage="special"), Schema()}

    suggestions = suggest_type("\\df myschema.xxx", "\\df myschema.xxx")
    assert suggestions == (Function(schema="myschema", usage="special"),)


def test_leading_whitespace_ok():
    cmd = "\\dn "
    whitespace = "   "
    suggestions = suggest_type(whitespace + cmd, whitespace + cmd)
    assert suggestions == suggest_type(cmd, cmd)


def test_dT_suggests_schema_or_datatypes():
    text = "\\dT "
    suggestions = suggest_type(text, text)
    assert set(suggestions) == {Schema(), Datatype(schema=None)}


def test_schema_qualified_dT_suggests_datatypes():
    text = "\\dT foo."
    suggestions = suggest_type(text, text)
    assert suggestions == (Datatype(schema="foo"),)


@pytest.mark.parametrize("command", ["\\c ", "\\connect "])
def test_c_suggests_databases(command):
    suggestions = suggest_type(command, command)
    assert suggestions == (Database(),)



import sys
import click
from .parseutils import is_destructive


def confirm_destructive_query(queries, keywords, alias):
    """Check if the query is destructive and prompts the user to confirm.

    Returns:
    * None if the query is non-destructive or we can't prompt the user.
    * True if the query is destructive and the user wants to proceed.
    * False if the query is destructive and the user doesn't want to proceed.

    """
    info = "You're about to run a destructive command"
    if alias:
        info += f" in {click.style(alias, fg='red')}"

    prompt_text = f"{info}.\nDo you want to proceed?"
    if is_destructive(queries, keywords) and sys.stdin.isatty():
        return confirm(prompt_text)


def confirm(*args, **kwargs):
    """Prompt for confirmation (yes/no) and handle any abort exceptions."""
    try:
        return click.confirm(*args, **kwargs)
    except click.Abort:
        return False


def prompt(*args, **kwargs):
    """Prompt the user for input and handle any abort exceptions."""
    try:
        return click.prompt(*args, **kwargs)
    except click.Abort:
        return False



"""Tests for SQL trailing comment handling.

Verifies that statements with comments after the semicolon are handled
correctly in both the input buffer (pgbuffer) and query execution (pgexecute).
"""

from pgcli.pgbuffer import _is_complete


class TestIsCompleteWithTrailingComments:
    """Test _is_complete() handles trailing SQL comments after semicolons."""

    def test_simple_semicolon(self):
        assert _is_complete("SELECT 1;") is True

    def test_no_semicolon(self):
        assert _is_complete("SELECT 1") is False

    def test_trailing_single_line_comment(self):
        assert _is_complete("SELECT 1; -- a comment") is True

    def test_trailing_block_comment(self):
        assert _is_complete("SELECT 1; /* block comment */") is True

    def test_vacuum_with_comment(self):
        assert _is_complete("vacuum freeze verbose tpd.file_delivery; -- 82% towards emergency") is True

    def test_comment_only(self):
        assert _is_complete("-- just a comment") is False

    def test_semicolon_inside_string(self):
        assert _is_complete("SELECT ';'") is False

    def test_semicolon_inside_string_with_trailing_comment(self):
        assert _is_complete("SELECT ';' FROM t; -- note") is True

    def test_open_quote(self):
        assert _is_complete("SELECT '") is False

    def test_empty_string(self):
        assert _is_complete("") is False

    def test_multiple_semicolons_with_comment(self):
        assert _is_complete("SELECT 1; SELECT 2; -- done") is True

    def test_comment_with_special_chars(self):
        assert _is_complete("VACUUM ANALYZE; -- 81.0% towards emergency, 971 MB") is True



import pytest
from click.testing import CliRunner

from pgcli.main import cli, PGCli


@pytest.fixture
def dummy_exec(monkeypatch, tmp_path):
    # Capture executed commands
    # Isolate config directory for tests
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    dummy_cmds = []

    class DummyExec:
        def run(self, cmd):
            # Ignore ping SELECT 1 commands used for exiting CLI
            if cmd.strip().upper() == "SELECT 1":
                return []
            # Record init commands
            dummy_cmds.append(cmd)
            return []

        def get_timezone(self):
            return "UTC"

        def set_timezone(self, *args, **kwargs):
            pass

    def fake_connect(self, *args, **kwargs):
        self.pgexecute = DummyExec()

    monkeypatch.setattr(PGCli, "connect", fake_connect)
    return dummy_cmds


def test_init_command_option(dummy_exec):
    "Test that --init-command triggers execution of the command."
    runner = CliRunner()
    # Use a custom init command and --ping to exit the CLI after init commands
    result = runner.invoke(cli, ["--init-command", "SELECT foo", "--ping", "db", "user"])
    assert result.exit_code == 0
    # Should print the init command
    assert "Running init commands: SELECT foo" in result.output
    # Should exit via ping
    assert "PONG" in result.output
    # DummyExec should have recorded only the init command
    assert dummy_exec == ["SELECT foo"]


def test_init_commands_from_config(dummy_exec, tmp_path):
    """
    Test that init commands defined in the config file are executed on startup.
    """
    # Create a temporary config file with init-commands
    config_file = tmp_path / "pgclirc_test"
    config_file.write_text("[main]\n[init-commands]\nfirst = SELECT foo;\nsecond = SELECT bar;\n")

    runner = CliRunner()
    # Use --ping to exit the CLI after init commands
    result = runner.invoke(cli, ["--pgclirc", str(config_file.absolute()), "--ping", "testdb", "user"])
    assert result.exit_code == 0
    # Should print both init commands in order (note trailing semicolons cause double ';;')
    assert "Running init commands: SELECT foo;; SELECT bar;" in result.output
    # DummyExec should have recorded both commands
    assert dummy_exec == ["SELECT foo;", "SELECT bar;"]


def test_init_commands_option_and_config(dummy_exec, tmp_path):
    """
    Test that CLI-provided init command is appended after config-defined commands.
    """
    # Create a temporary config file with init-commands
    config_file = tmp_path / "pgclirc_test"
    config_file.write_text("[main]\n [init-commands]\nfirst = SELECT foo;\n")

    runner = CliRunner()
    # Use --ping to exit the CLI after init commands
    result = runner.invoke(
        cli,
        [
            "--pgclirc",
            str(config_file),
            "--init-command",
            "SELECT baz;",
            "--ping",
            "testdb",
            "user",
        ],
    )
    assert result.exit_code == 0
    # Should print config command followed by CLI option (double ';' between commands)
    assert "Running init commands: SELECT foo;; SELECT baz;" in result.output
    # DummyExec should record both commands in order
    assert dummy_exec == ["SELECT foo;", "SELECT baz;"]



from pgcli.pyev import Visualizer
import json


"""Explain response output adapter"""


class ExplainOutputFormatter:
    def __init__(self, max_width):
        self.max_width = max_width

    def format_output(self, cur, headers, **output_kwargs):
        # explain query results should always contain 1 row each
        [(data,)] = list(cur)
        explain_list = json.loads(data)
        visualizer = Visualizer(self.max_width)
        for explain in explain_list:
            visualizer.load(explain)
            yield visualizer.get_list()



"""
Steps for behavioral style tests are defined in this module.
Each step is defined by the string decorating it.
This string is used to call the step in "*.feature" file.
"""

from behave import when, then
import wrappers


@when('we send "show help" command')
def step_send_help_command(context):
    context.cli.sendline("show help")


@then("we see the pgbouncer help output")
def see_pgbouncer_help(context):
    wrappers.expect_exact(
        context,
        "SHOW HELP|CONFIG|DATABASES|POOLS|CLIENTS|SERVERS|USERS|VERSION",
        timeout=3,
    )



import textwrap
import re
from click import style as color

DESCRIPTIONS = {
    "Append": "Used in a UNION to merge multiple record sets by appending them together.",
    "Limit": "Returns a specified number of rows from a record set.",
    "Sort": "Sorts a record set based on the specified sort key.",
    "Nested Loop": "Merges two record sets by looping through every record in the first set and trying to find a match in the second set. All matching records are returned.",
    "Merge Join": "Merges two record sets by first sorting them on a join key.",
    "Hash": "Generates a hash table from the records in the input recordset. Hash is used by Hash Join.",
    "Hash Join": "Joins to record sets by hashing one of them (using a Hash Scan).",
    "Aggregate": "Groups records together based on a GROUP BY or aggregate function (e.g. sum()).",
    "Hashaggregate": "Groups records together based on a GROUP BY or aggregate function (e.g. sum()). Hash Aggregate uses a hash to first organize the records by a key.",
    "Sequence Scan": "Finds relevant records by sequentially scanning the input record set. When reading from a table, Seq Scans (unlike Index Scans) perform a single read operation (only the table is read).",
    "Seq Scan": "Finds relevant records by sequentially scanning the input record set. When reading from a table, Seq Scans (unlike Index Scans) perform a single read operation (only the table is read).",
    "Index Scan": "Finds relevant records based on an Index. Index Scans perform 2 read operations: one to read the index and another to read the actual value from the table.",
    "Index Only Scan": "Finds relevant records based on an Index. Index Only Scans perform a single read operation from the index and do not read from the corresponding table.",
    "Bitmap Heap Scan": "Searches through the pages returned by the Bitmap Index Scan for relevant rows.",
    "Bitmap Index Scan": "Uses a Bitmap Index (index which uses 1 bit per page) to find all relevant pages. Results of this node are fed to the Bitmap Heap Scan.",
    "CTEScan": "Performs a sequential scan of Common Table Expression (CTE) query results. Note that results of a CTE are materialized (calculated and temporarily stored).",
    "ProjectSet": "ProjectSet appears when the SELECT or ORDER BY clause of the query.  They basically just execute the set-returning function(s) for each tuple until none of the functions return any more records.",
    "Result": "Returns result",
}


class Visualizer:
    def __init__(self, terminal_width=100, color=True):
        self.color = color
        self.terminal_width = terminal_width
        self.string_lines = []

    def load(self, explain_dict):
        self.plan = explain_dict.pop("Plan")
        self.explain = explain_dict
        self.process_all()
        self.generate_lines()

    def process_all(self):
        self.plan = self.process_plan(self.plan)
        self.plan = self.calculate_outlier_nodes(self.plan)

    #
    def process_plan(self, plan):
        plan = self.calculate_planner_estimate(plan)
        plan = self.calculate_actuals(plan)
        self.calculate_maximums(plan)
        #
        for index in range(len(plan.get("Plans", []))):
            _plan = plan["Plans"][index]
            plan["Plans"][index] = self.process_plan(_plan)
        return plan

    def prefix_format(self, v):
        if self.color:
            return color(v, fg="bright_black")
        return v

    def tag_format(self, v):
        if self.color:
            return color(v, fg="white", bg="red")
        return v

    def muted_format(self, v):
        if self.color:
            return color(v, fg="bright_black")
        return v

    def bold_format(self, v):
        if self.color:
            return color(v, fg="white")
        return v

    def good_format(self, v):
        if self.color:
            return color(v, fg="green")
        return v

    def warning_format(self, v):
        if self.color:
            return color(v, fg="yellow")
        return v

    def critical_format(self, v):
        if self.color:
            return color(v, fg="red")
        return v

    def output_format(self, v):
        if self.color:
            return color(v, fg="cyan")
        return v

    def calculate_planner_estimate(self, plan):
        plan["Planner Row Estimate Factor"] = 0
        plan["Planner Row Estimate Direction"] = "Under"

        if plan["Plan Rows"] == plan["Actual Rows"]:
            return plan

        if plan["Plan Rows"] != 0:
            plan["Planner Row Estimate Factor"] = plan["Actual Rows"] / plan["Plan Rows"]

        if plan["Planner Row Estimate Factor"] < 10:
            plan["Planner Row Estimate Factor"] = 0
            plan["Planner Row Estimate Direction"] = "Over"
            if plan["Actual Rows"] != 0:
                plan["Planner Row Estimate Factor"] = plan["Plan Rows"] / plan["Actual Rows"]
        return plan

    #
    def calculate_actuals(self, plan):
        plan["Actual Duration"] = plan["Actual Total Time"]
        plan["Actual Cost"] = plan["Total Cost"]

        for child in plan.get("Plans", []):
            if child["Node Type"] != "CTEScan":
                plan["Actual Duration"] = plan["Actual Duration"] - child["Actual Total Time"]
                plan["Actual Cost"] = plan["Actual Cost"] - child["Total Cost"]

        if plan["Actual Cost"] < 0:
            plan["Actual Cost"] = 0

        plan["Actual Duration"] = plan["Actual Duration"] * plan["Actual Loops"]
        return plan

    def calculate_outlier_nodes(self, plan):
        plan["Costliest"] = plan["Actual Cost"] == self.explain["Max Cost"]
        plan["Largest"] = plan["Actual Rows"] == self.explain["Max Rows"]
        plan["Slowest"] = plan["Actual Duration"] == self.explain["Max Duration"]

        for index in range(len(plan.get("Plans", []))):
            _plan = plan["Plans"][index]
            plan["Plans"][index] = self.calculate_outlier_nodes(_plan)
        return plan

    def calculate_maximums(self, plan):
        if not self.explain.get("Max Rows"):
            self.explain["Max Rows"] = plan["Actual Rows"]
        elif self.explain.get("Max Rows") < plan["Actual Rows"]:
            self.explain["Max Rows"] = plan["Actual Rows"]

        if not self.explain.get("Max Cost"):
            self.explain["Max Cost"] = plan["Actual Cost"]
        elif self.explain.get("Max Cost") < plan["Actual Cost"]:
            self.explain["Max Cost"] = plan["Actual Cost"]

        if not self.explain.get("Max Duration"):
            self.explain["Max Duration"] = plan["Actual Duration"]
        elif self.explain.get("Max Duration") < plan["Actual Duration"]:
            self.explain["Max Duration"] = plan["Actual Duration"]

        if not self.explain.get("Total Cost"):
            self.explain["Total Cost"] = plan["Actual Cost"]
        elif self.explain.get("Total Cost") < plan["Actual Cost"]:
            self.explain["Total Cost"] = plan["Actual Cost"]

    #
    def duration_to_string(self, value):
        if value < 1:
            return self.good_format("<1 ms")
        elif value < 100:
            return self.good_format("%.2f ms" % value)
        elif value < 1000:
            return self.warning_format("%.2f ms" % value)
        elif value < 60000:
            return self.critical_format(
                "%.2f s" % (value / 1000.0),
            )
        else:
            return self.critical_format(
                "%.2f m" % (value / 60000.0),
            )

    # }
    #
    def format_details(self, plan):
        details = []

        if plan.get("Scan Direction"):
            details.append(plan["Scan Direction"])

        if plan.get("Strategy"):
            details.append(plan["Strategy"])

        if len(details) > 0:
            return self.muted_format(" [%s]" % ", ".join(details))

        return ""

    def format_tags(self, plan):
        tags = []

        if plan["Slowest"]:
            tags.append(self.tag_format("slowest"))
        if plan["Costliest"]:
            tags.append(self.tag_format("costliest"))
        if plan["Largest"]:
            tags.append(self.tag_format("largest"))
        if plan.get("Planner Row Estimate Factor", 0) >= 100:
            tags.append(self.tag_format("bad estimate"))

        return " ".join(tags)

    def get_terminator(self, index, plan):
        if index == 0:
            if len(plan.get("Plans", [])) == 0:
                return "⌡► "
            else:
                return "├►  "
        else:
            if len(plan.get("Plans", [])) == 0:
                return "   "
            else:
                return "│  "

    def wrap_string(self, line, width):
        if width == 0:
            return [line]
        return textwrap.wrap(line, width)

    def intcomma(self, value):
        sep = ","
        if not isinstance(value, str):
            value = int(value)

        orig = str(value)

        new = re.sub(r"^(-?\d+)(\d{3})", rf"\g<1>{sep}\g<2>", orig)
        if orig == new:
            return new
        else:
            return self.intcomma(new)

    def output_fn(self, current_prefix, string):
        return "%s%s" % (self.prefix_format(current_prefix), string)

    def create_lines(self, plan, prefix, depth, width, last_child):
        current_prefix = prefix
        self.string_lines.append(self.output_fn(current_prefix, self.prefix_format("│")))

        joint = "├"
        if last_child:
            joint = "└"
        #
        self.string_lines.append(
            self.output_fn(
                current_prefix,
                "%s %s%s %s"
                % (
                    self.prefix_format(joint + "─⌠"),
                    self.bold_format(plan["Node Type"]),
                    self.format_details(plan),
                    self.format_tags(plan),
                ),
            )
        )
        #
        if last_child:
            prefix += "  "
        else:
            prefix += "│ "

        current_prefix = prefix + "│ "

        cols = width - len(current_prefix)

        for line in self.wrap_string(
            DESCRIPTIONS.get(plan["Node Type"], "Not found : %s" % plan["Node Type"]),
            cols,
        ):
            self.string_lines.append(self.output_fn(current_prefix, "%s" % self.muted_format(line)))
        #
        if plan.get("Actual Duration"):
            self.string_lines.append(
                self.output_fn(
                    current_prefix,
                    "○ %s %s (%.0f%%)"
                    % (
                        "Duration:",
                        self.duration_to_string(plan["Actual Duration"]),
                        (plan["Actual Duration"] / self.explain["Execution Time"]) * 100,
                    ),
                )
            )

        self.string_lines.append(
            self.output_fn(
                current_prefix,
                "○ %s %s (%.0f%%)"
                % (
                    "Cost:",
                    self.intcomma(plan["Actual Cost"]),
                    (plan["Actual Cost"] / self.explain["Total Cost"]) * 100,
                ),
            )
        )

        self.string_lines.append(
            self.output_fn(
                current_prefix,
                "○ %s %s" % ("Rows:", self.intcomma(plan["Actual Rows"])),
            )
        )

        current_prefix = current_prefix + "  "

        if plan.get("Join Type"):
            self.string_lines.append(
                self.output_fn(
                    current_prefix,
                    "%s %s" % (plan["Join Type"], self.muted_format("join")),
                )
            )

        if plan.get("Relation Name"):
            self.string_lines.append(
                self.output_fn(
                    current_prefix,
                    "%s %s.%s"
                    % (
                        self.muted_format("on"),
                        plan.get("Schema", "unknown"),
                        plan["Relation Name"],
                    ),
                )
            )

        if plan.get("Index Name"):
            self.string_lines.append(
                self.output_fn(
                    current_prefix,
                    "%s %s" % (self.muted_format("using"), plan["Index Name"]),
                )
            )

        if plan.get("Index Condition"):
            self.string_lines.append(
                self.output_fn(
                    current_prefix,
                    "%s %s" % (self.muted_format("condition"), plan["Index Condition"]),
                )
            )

        if plan.get("Filter"):
            self.string_lines.append(
                self.output_fn(
                    current_prefix,
                    "%s %s %s"
                    % (
                        self.muted_format("filter"),
                        plan["Filter"],
                        self.muted_format("[-%s rows]" % self.intcomma(plan["Rows Removed by Filter"])),
                    ),
                )
            )

        if plan.get("Hash Condition"):
            self.string_lines.append(
                self.output_fn(
                    current_prefix,
                    "%s %s" % (self.muted_format("on"), plan["Hash Condition"]),
                )
            )

        if plan.get("CTE Name"):
            self.string_lines.append(self.output_fn(current_prefix, "CTE %s" % plan["CTE Name"]))

        if plan.get("Planner Row Estimate Factor") != 0:
            self.string_lines.append(
                self.output_fn(
                    current_prefix,
                    "%s %sestimated %s %.2fx"
                    % (
                        self.muted_format("rows"),
                        plan["Planner Row Estimate Direction"],
                        self.muted_format("by"),
                        plan["Planner Row Estimate Factor"],
                    ),
                )
            )

        current_prefix = prefix

        if len(plan.get("Output", [])) > 0:
            for index, line in enumerate(self.wrap_string(" + ".join(plan["Output"]), cols)):
                self.string_lines.append(
                    self.output_fn(
                        current_prefix,
                        self.prefix_format(self.get_terminator(index, plan)) + self.output_format(line),
                    )
                )

        for index, nested_plan in enumerate(plan.get("Plans", [])):
            self.create_lines(nested_plan, prefix, depth + 1, width, index == len(plan["Plans"]) - 1)

    def generate_lines(self):
        self.string_lines = [
            "○ Total Cost: %s" % self.intcomma(self.explain["Total Cost"]),
            "○ Planning Time: %s" % self.duration_to_string(self.explain["Planning Time"]),
            "○ Execution Time: %s" % self.duration_to_string(self.explain["Execution Time"]),
            self.prefix_format("┬"),
        ]
        self.create_lines(
            self.plan,
            "",
            0,
            self.terminal_width,
            len(self.plan.get("Plans", [])) == 1,
        )

    def get_list(self):
        return "\n".join(self.string_lines)

    def print(self):
        for lin in self.string_lines:
            print(lin)



#!/usr/bin/env python
"""A script to publish a release of pgcli to PyPI."""

import io
import re
import subprocess
import sys
from optparse import OptionParser

import click

DEBUG = False
CONFIRM_STEPS = False
DRY_RUN = False


def skip_step():
    """
    Asks for user's response whether to run a step. Default is yes.
    :return: boolean
    """
    global CONFIRM_STEPS

    if CONFIRM_STEPS:
        return not click.confirm("--- Run this step?", default=True)
    return False


def run_step(*args):
    """
    Prints out the command and asks if it should be run.
    If yes (default), runs it.
    :param args: list of strings (command and args)
    """
    global DRY_RUN

    cmd = args
    print(" ".join(cmd))
    if skip_step():
        print("--- Skipping...")
    elif DRY_RUN:
        print("--- Pretending to run...")
    else:
        subprocess.check_output(cmd)


def version(version_file):
    _version_re = re.compile(r'__version__\s+=\s+(?P<quote>[\'"])(?P<version>.*)(?P=quote)')

    with io.open(version_file, encoding="utf-8") as f:
        ver = _version_re.search(f.read()).group("version")

    return ver


def get_merged_prs_since_last_tag():
    """Get list of PR numbers and titles merged since the last tag."""
    try:
        previous_tag = (
            subprocess
            .check_output(
                ["git", "describe", "--abbrev=0", "--tags"],
                stderr=subprocess.DEVNULL,
            )
            .decode()
            .strip()
        )
    except subprocess.CalledProcessError:
        return []

    log = subprocess.check_output(["git", "log", "--merges", "--oneline", "{}..HEAD".format(previous_tag)]).decode()

    prs = re.findall(r"(.+\(#(\d+)\))", log)
    seen = set()
    result = []
    for line, num in prs:
        if num not in seen:
            seen.add(num)
            result.append(num)
    return result


def commit_for_release(version_file, ver):
    pr_numbers = get_merged_prs_since_last_tag()
    pr_list = ""
    if pr_numbers:
        pr_list = "\n\n" + "\n".join("- #{}".format(n) for n in pr_numbers)

    message = "Releasing version {}{}".format(ver, pr_list)
    run_step("git", "reset")
    run_step("git", "add", "-u")
    run_step("git", "commit", "--message", message)


def create_git_tag(tag_name):
    run_step("git", "tag", tag_name)


def create_distribution_files():
    run_step("rm", "-rf", "dist/")
    run_step("python", "-m", "build")


def upload_distribution_files():
    run_step("twine", "upload", "dist/*")


def push_to_github():
    run_step("git", "push", "origin", "main")


def push_tags_to_github():
    run_step("git", "push", "--tags", "origin")


def check_tag(ver):
    """Verify that HEAD is on the expected tag."""
    tag = "v{}".format(ver)
    try:
        current_tag = (
            subprocess
            .check_output(
                ["git", "describe", "--exact-match", "--tags", "HEAD"],
                stderr=subprocess.DEVNULL,
            )
            .decode()
            .strip()
        )
    except subprocess.CalledProcessError:
        print("ERROR: HEAD is not on any tag. Expected tag '{}'.".format(tag))
        sys.exit(1)
    if current_tag != tag:
        print("ERROR: HEAD is on tag '{}', expected '{}'.".format(current_tag, tag))
        sys.exit(1)
    print("OK: on tag '{}'".format(tag))


def comment_on_released_prs(ver):
    """Post a comment on all PRs included in this release."""
    tag = "v{}".format(ver)
    try:
        previous_tag = (
            subprocess
            .check_output(
                ["git", "describe", "--abbrev=0", "--tags", "{}^".format(tag)],
                stderr=subprocess.DEVNULL,
            )
            .decode()
            .strip()
        )
    except subprocess.CalledProcessError:
        print("WARNING: Could not find previous tag. Skipping PR comments.")
        return

    log = subprocess.check_output(["git", "log", "--merges", "--oneline", "{}..{}".format(previous_tag, tag)]).decode()

    pr_numbers = re.findall(r"#(\d+)", log)
    pr_numbers = list(set(pr_numbers))

    if not pr_numbers:
        print("No PRs found between {} and {}.".format(previous_tag, tag))
        return

    print("Found PRs: {}".format(", ".join("#" + n for n in pr_numbers)))
    message = "Released as part of {}.".format(ver)

    print("gh pr comment --body '{}' {}".format(message, " ".join(pr_numbers)))
    if skip_step():
        print("--- Skipping...")
    elif DRY_RUN:
        print("--- Pretending to run...")
    else:
        for pr in pr_numbers:
            subprocess.check_output(["gh", "pr", "comment", pr, "--body", message])


def checklist(questions):
    for question in questions:
        if not click.confirm("--- {}".format(question), default=False):
            sys.exit(1)


if __name__ == "__main__":
    if DEBUG:
        subprocess.check_output = lambda x: x

    # checks = [
    #     "Have you updated the AUTHORS file?",
    #     "Have you updated the `Usage` section of the README?",
    # ]
    # checklist(checks)

    ver = version("pgcli/__init__.py")
    print("Releasing Version:", ver)

    parser = OptionParser()
    parser.add_option(
        "-c",
        "--confirm-steps",
        action="store_true",
        dest="confirm_steps",
        default=False,
        help=("Confirm every step. If the step is not confirmed, it will be skipped."),
    )
    parser.add_option(
        "-d",
        "--dry-run",
        action="store_true",
        dest="dry_run",
        default=False,
        help="Print out, but not actually run any steps.",
    )

    popts, pargs = parser.parse_args()
    CONFIRM_STEPS = popts.confirm_steps
    DRY_RUN = popts.dry_run

    if not click.confirm("Are you sure?", default=False):
        sys.exit(1)

    commit_for_release("pgcli/__init__.py", ver)
    create_git_tag("v{}".format(ver))
    push_to_github()
    push_tags_to_github()
    check_tag(ver)
    create_distribution_files()
    upload_distribution_files()
    comment_on_released_prs(ver)



from pgcli.packages.parseutils.meta import FunctionMetadata


def test_function_metadata_eq():
    f1 = FunctionMetadata("s", "f", ["x"], ["integer"], [], "int", False, False, False, False, None)
    f2 = FunctionMetadata("s", "f", ["x"], ["integer"], [], "int", False, False, False, False, None)
    f3 = FunctionMetadata("s", "g", ["x"], ["integer"], [], "int", False, False, False, False, None)
    assert f1 == f2
    assert f1 != f3
    assert not (f1 != f2)
    assert not (f1 == f3)
    assert hash(f1) == hash(f2)
    assert hash(f1) != hash(f3)
